﻿(*
  .exe .dll の埋め込みバージョン情報を表示する
  Delphi 11 用ソース

  Ver1.06 2023/07/20 Win64対策を行った
*)

// Debugモードの場合 ExeVer.exe の有るフォルダにダンプファイルを作成する
{$IFDEF DEBUG}
  {$DEFINE BinDump}    // バイナリー・ダンプ
  {$DEFINE HexDump}    // ＨＥＸダンプ
{$ENDIF}


unit Main;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, ShellAPI, MultiMon, Clipbrd,
  Vcl.Grids, Vcl.ComCtrls, ComObj, ShlObj, ActiveX, System.UITypes, AnsiStrings,
  Vcl.Mask;

type
  TFormMain = class(TForm)
    ButtonOpen: TButton;
    LabeledEditFileName: TLabeledEdit;
    OpenDialog1: TOpenDialog;
    ListView1: TListView;
    ButtonExit: TButton;
    ButtonCopyCom: TButton;
    ButtonCopyTab: TButton;
    CheckBoxBlank: TCheckBox;
    procedure ButtonOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure OpenDialog1Show(Sender: TObject);
    procedure OpenDialog1Close(Sender: TObject);
    procedure ButtonExitClick(Sender: TObject);
    procedure LabeledEditFileNameKeyPress(Sender: TObject; var Key: Char);
    procedure FormResize(Sender: TObject);
    procedure ButtonCopyComClick(Sender: TObject);
    procedure ButtonCopyTabClick(Sender: TObject);
    procedure CheckBoxBlankClick(Sender: TObject);
  private
    { Private 宣言 }
    FOpenDialog1hWnd: HWND;
    FNewOpenDialog1WndProc, FOldOpenDialog1WndProc: TFarProc;
    FOpenDialog1MoveFlag: Byte;
    procedure OpenDialog1WndProc(var Message: TMessage);
    Procedure WMDropFiles(Var Msg : TMessage); Message WM_DROPFILES;   // for Drag&Drop
  public
    { Public 宣言 }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.dfm}


  // フォームTargetを表示しているモニタの座標とサイズを返す（Uses に MultiMon が必要）
  Function GetMonitorRect(Target : TForm = Nil) : TRect;
  Var LMonitorInfoEx : TMonitorInfoEx;
  begin
    FillChar(LMonitorInfoEx, SizeOf(TMonitorInfoEx), #0);
    LMonitorInfoEx.cbSize := SizeOf(LMonitorInfoEx);
    if Target=Nil then Target := Application.MainForm;
    GetMonitorInfo(Target.Monitor.Handle, @LMonitorInfoEx);
    Result := LMonitorInfoEx.rcWork;     // 有効エリアを返す
//  Result := LMonitorInfoEx.rcMonitor;  // フルエリアを返す場合
  end;


  //----------------------------- for FormMain.OpenDialog1 -----------------------------------------
  //                    OpenDialog を メインフォームの中央に表示する

  // OpenDialog1 OnShowイベント
  procedure TFormMain.OpenDialog1Show(Sender: TObject);
  begin
    if ofOldStyleDialog in (Sender as TOpenDialog).Options then
      FOpenDialog1hWnd := (Sender as TOpenDialog).Handle
    else
      FOpenDialog1hWnd := GetParent((Sender as TOpenDialog).Handle);
    // 動的サブクラス化
    FOpenDialog1MoveFlag := 3; // 位置調整していない
    FNewOpenDialog1WndProc := MakeObjectInstance(OpenDialog1WndProc);
    FOldOpenDialog1WndProc := Pointer(GetWindowLong(FOpenDialog1hWnd, GWL_WNDPROC));
    SetWindowLong(FOpenDialog1hWnd, GWL_WNDPROC, Int64(FNewOpenDialog1WndProc));
  end;

  // OpenDialog1 OnCloseイベント
  procedure TFormMain.OpenDialog1Close(Sender: TObject);
  begin
    // 動的サブクラス化を解除する
    SetWindowLong(FOpenDialog1hWnd, GWL_WNDPROC, INT64(FOldOpenDialog1WndProc));
    FreeObjectInstance(FNewOpenDialog1WndProc);
    FOpenDialog1hWnd       := HWND(nil);
    FNewOpenDialog1WndProc := nil;
    FOldOpenDialog1WndProc := nil;
  end;

  // OpenDialog1の新しいウィンドウプロシージャ
  procedure TFormMain.OpenDialog1WndProc(var Message: TMessage);
  var
    pWinPos: PWindowPos;
    ScreenRect, mc, rc: TRect;
  begin
    if (Message.Msg = WM_WINDOWPOSCHANGING) and (FOpenDialog1MoveFlag > 0) then begin
      pWinPos := PWindowPos(Message.LParam);
      pWinPos^.flags := pWinPos^.flags and not(SWP_NOMOVE);
      Dec(FOpenDialog1MoveFlag);
      // 表示位置をスクリーン座標で設定する
      ScreenRect := GetMonitorRect(FormMain);
      GetWindowRect(FormMain.Handle, mc);
      GetWindowRect(FOpenDialog1hWnd, rc);
      pWinPos^.x := mc.left + (mc.Width  - rc.Width)  div 2;
      pWinPos^.y := mc.Top  + (mc.Height - rc.Height) div 2;
      // スクリーンからはみ出さないように制限する
      if pWinPos^.y<ScreenRect.Top then pWinPos^.y := ScreenRect.Top;
      if pWinPos^.y>(ScreenRect.Bottom-rc.Height) then pWinPos^.y := ScreenRect.Bottom-rc.Height;
      if pWinPos^.x<ScreenRect.Left then pWinPos^.x := ScreenRect.Left;
      if pWinPos^.x>(ScreenRect.Right-rc.Width) then pWinPos^.x := ScreenRect.Right-rc.Width;
    end;
    Message.Result := CallWindowProc(FOldOpenDialog1WndProc, FOpenDialog1hWnd, Message.Msg, Message.WParam, Message.LParam);
  end;

  //-------------------------------------------------------------------

  // 実行中のファイルのバージョン情報を取り出す
  function GetFxeVersion : String;
  var
    aVersion : Cardinal;
    St : String;
  begin
    aVersion := GetFileVersion(Application.ExeName);  //ファイルのバージョンを取得
    St := IntTostr(aVersion and $FFFF);
    while Length(St)<2 do St := '0' + St;
    Result := IntToStr(aVersion shr 16) +'.'+ St;
  end;

  //-------------------------------------------------------------------


// リンクファイルのファイル名を取り出す
// Usesに  ComObj, ShlObj, ActiveX が必要
function GetLnkFname(Fn : String) : String;
var
  ShellLink     : IShellLink;
  PersistFile   : IPersistFile;
  Win32FindData : TWin32FindData;
  S, SS : String;

  Function Get81(WFilename : String) : String;
  var
    Fs : File of Byte;
    Buff : Array[0..2000] of byte;
    Rb, n, P : Integer;
    St : String;
  begin
    try
      Assign(Fs, WFilename); Reset(Fs);
      BlockRead(Fs, Buff, Sizeof(Buff), Rb);
      CloseFile(Fs);
      if Buff[$14]=$81 then begin
        St := ''; n := 0; P := $A1;
        While ((Buff[P+n*2+1] SHL 8 + Buff[P+n*2])<>0) and (n<Max_Path) do begin
          St := St+Chr(Buff[P+n*2+1] SHL 8 + Buff[P+n*2]);
          Inc(n);
        end;
        try
          if Not FileExists(St) then begin
            P := $78+Buff[$78];
            if P<Rb then P := P+Buff[P]+(Buff[P+1] SHL 8);
            if P<Rb then P := P+Buff[P]+(Buff[P+1] SHL 8);
            if P<Rb then P := P+Buff[P]+(Buff[P+1] SHL 8);
            if P<Rb then P := P+Buff[P]+(Buff[P+1] SHL 8);
            if P<Rb then P := P+$29;
            St := ''; n := 0;
            if P<=(Rb+n*2) then begin
              While ((Buff[P+n*2+1] SHL 8 + Buff[P+n*2])<>0) and (n<Max_Path) do begin
                St := St+Chr(Buff[P+n*2+1] SHL 8 + Buff[P+n*2]);
                Inc(n);
              end;
              if Not FileExists(St) then St := '';
            end;
          end;
        except
          St := '';
        end;
        Result := St;
      end
      else
        Result := '';
    except
      Result := '';
    end;
  end;

begin
  ShellLink := CreateComObject(CLSID_ShellLink) as IShellLink;
  PersistFile := ShellLink as IPersistFile;
  if Succeeded(PerSistFile.Load(PWChar(Fn), STGM_READ)) then begin
//  ShellLink.Resolve(0, SLR_ANY_MATCH);
    ShellLink.Resolve(0, SLR_NO_UI);
    SetLength(S, MAX_PATH);
    //ショートカットファイルの参照先を取得
    ShellLink.GetPath(PChar(S), MAX_PATH, Win32FindData, SLGP_UNCPRIORITy);
    if Pos(#0, S)>0 then S := Copy(S, 1, Pos(#0, S)-1);
    if Not FileExists(S) and (Pos('\Program Files (x86)\', S)>=1) then begin
      SS := StringReplace(S, '\Program Files (x86)\', '\Program Files\', []);
      if FileExists(SS) then S := SS;
    end;
    if S='' then S := Get81(Fn);
    Result := S;
  end;
end;

//-------------------------------------------------------------------


// バージョン情報を取得する Item<TAB>Content
function GetVersionName(Fn : String): string;
const
  Translation = '\VarFileInfo\Translation';
var
  BufSize : Integer;
  HWnd : DWORD;
  VerInfoBuf : Array of Char;

  // 2023/7/20 For Win64
  Function FindKey(Key : String; Var Buf : Array of Char) : Integer;
  var n, m : Integer;
  begin
    m := 0;
    Repeat
      n := 1;
      while (m<=Length(Buf)) and (n<=Length(Key)) and (Buf[m+n]=Key[n]) do inc(n);
      Inc(m);
    until ((n+m)>Length(Buf)) or (n>=Length(Key));
    if n>=Length(Key) then Result := m else Result := 0;
  end;

  // バージョン情報を文字列に変換する
  Function GetName : String;
  Var
    Pt, Bt, Sp, Ds, Sl : Integer;
    St : String;
{$IFDEF BinDump}
    F: File of Char;
{$ENDIF}
{$IFDEF HexDump}
    Ss : String; Ts : TStrings; m : Integer;
{$ENDIF}
  begin
    Ds := Ord(VerInfoBuf[0]) div 2;     // 有効データサイズ(Word)

{$IFDEF BinDump}
    // バイナリーダンプ
    Assign(F, ExtractFilePath(Application.ExeName)+ExtractFileName(Fn)+'.dat'); Rewrite(F); BlockWrite(F, VerInfoBuf[0], Ds+2); CloseFile(F);
{$ENDIF}

{$IFDEF HexDump}
    // HEXダンプ
    St := Fn; Ss := ''; m := 0;
    for Pt := 0 to Ds+1 do begin
      m := Pt Mod 16;
      if m=0 then begin
        St := St + '  '+ Ss + sLineBreak;
        St := St + IntToHex(Pt, 4) + ':';
        Ss := '';
      end;
      St := St + ' ' + IntToHex(Ord(VerInfoBuf[Pt]), 4);
      if VerInfoBuf[Pt]>=' ' then Ss := Ss + VerInfoBuf[Pt] else Ss := Ss + '~';
    end;
    if m<>0 then
      St := St + StringOfChar(' ', (15-m)*5) + '  '+ Ss + sLineBreak;
    Ts := TstringList.Create;
    try
      Ts.Text := St;
      Ts.SaveToFile(ExtractFilePath(Application.ExeName)+ExtractFileName(Fn)+'.txt');
    Finally
      Ts.Free;
    end;
{$ENDIF}

    St := '';
    Result := '';

    Pt := FindKey('StringFileInfo'#0, VerInfoBuf);
//  Pt := Pos('StringFileInfo'#0, String(VerInfoBuf)); // 先頭位置を探索する Win64ではNG
    if Pt>0 then begin
      Pt := Pt + $1A;
      while (Pt<Ds-1) do begin
        Bt := Ord(VerInfoBuf[Pt]) div 2;     // ブロックサイズ
        Sl := Ord(VerInfoBuf[Pt+1]) div 2;   // 文字長
        while (Bt=0) and (Pt<Ds) do begin
          Inc(Pt);
          Bt := Ord(VerInfoBuf[Pt]) div 2;   // ブロックサイズ
          Sl := Ord(VerInfoBuf[Pt+1]) div 2; // 文字長
        end;
        Sp := Pt+3;
        while (VerInfoBuf[Sp]<>#0) and (Pt<Ds) do begin        // Item文字列取出し
          St := St + VerInfoBuf[Sp];
          Inc(Sp);
        end;
        St := St +^I;
        if Sl>0 then begin
          while (VerInfoBuf[Sp]=#0) and (Pt<Ds) do Sp := Sp+1; // Null文字をスキップ
          while (VerInfoBuf[Sp]<>#0) and (Pt<Ds) do begin      // Content文字列取出し
            St := St + VerInfoBuf[Sp];
            Inc(Sp);
          end;
        end;
        St := St + sLineBreak;
        Pt := Pt + Bt;
      end;
      Result := St;
    end;
  end;

begin
  Result := '';
  BufSize := GetFileVersionInfoSize(PChar(Fn), HWnd);           // 必要なバッファのサイズを取得
  if (BufSize<>0) then begin
    SetLength(VerInfoBuf, BufSize div 2);                       // バッファサイズを設定する（BufSize:1Byte単位, VerInfoBuf:2Byte単位）
    GetFileVersionInfo(PChar(Fn), 0, BufSize, @VerInfoBuf[0]);  // バッファにバージョン情報を取り出す
    Result := GetName;                                          // バージョン情報を文字列に変換する
    Finalize(VerInfoBuf);
  end;
end;


// .exe .dll の 32bit版と64bit版を判定する 2021/12/06
Function Get32or64(Fn :String) : String;
Const
  Pre = AnsiString(#$50#$45#$00#$00);
  C32 = AnsiString(#$4C#$01);
  C64 = AnsiString(#$64#$86);
Var
  n, Dlen : Integer;
  Buf : Array[1..$300] of AnsiChar;
  Fs : TFileStream;
begin
  Result := '';
  Try
  Fs := TFileStream.Create(Fn, fmOpenRead);
    Try
      Dlen := Fs.Read(Buf, SizeOf(Buf));
      n := 1;
      while (n<(Dlen-6)) and (Result='') do begin
        if (Buf[n]=Pre[1]) and (Buf[n+1]=Pre[2]) and (Buf[n+2]=Pre[3]) and (Buf[n+3]=Pre[4]) then begin
          if (Buf[n+4]=C32[1]) and (Buf[n+5]=C32[2]) then Result := '32bit';
          if (Buf[n+4]=C64[1]) and (Buf[n+5]=C64[2]) then Result := '64bit';
        end;
        Inc(n);
      end;
    Finally
      Fs.Free;
    End;
  Except
    Result := 'Unable to read : ' + ExtractFilename(Fn);
  End;
end;


// ファイル情報を表示する
Procedure DispInfo(Fn : String);
Var i, nn : Integer;
  KeyWordStr : TStrings;
  St, Cst : String;
  ListItem: TListItem;

  function TrimN(St : String) : String;
  begin
    if Not FormMain.CheckBoxBlank.Checked then St := Trim(St);  // 文字列前後の非表示文字を先所する
    Result := St;
  end;

begin
  With FormMain do begin
    KeyWordStr := TStringList.Create;
    try
      KeyWordStr.Text := GetVersionName(Fn); // バージョン情報を取得
      ListView1.Clear;
      // バージョン情報を表示する
      for i := 0 to KeyWordStr.Count-1 do begin
        St := KeyWordStr[i];
        if St <>'' Then begin
          nn := Pos(^I, St);
          Cst := TrimN(Copy(St, nn+1, Length(St)-nn));
          if CheckBoxBlank.Checked or (Cst<>'') then begin
            ListItem := ListView1.Items.Add;
            ListItem.Caption := Copy(St, 1, nn-1);
            ListItem.SubItems.Add(Cst);
          end;
        end;
      end;
    finally
      KeyWordStr.Free;
    end;

    // 2021/12/06
    St := Get32or64(Fn);
    if St<>'' then begin
      ListItem := ListView1.Items.Add;
      ListItem := ListView1.Items.Add;
      ListItem.Caption := 'bits';
      ListItem.SubItems.Add(St);
    end;

  end;
end;


// OpeDialogを開いてファイルを選択する
procedure TFormMain.ButtonOpenClick(Sender: TObject);
begin
  if OpenDialog1.Execute then begin
    LabeledEditFileName.Text := OpenDialog1.FileName;
    if LowerCase(ExtractFileExt(LabeledEditFileName.Text)) = '.lnk' then
      LabeledEditFileName.Text := GetLnkFname(LabeledEditFileName.Text);
    DispInfo(LabeledEditFileName.Text);
  end;
end;


// Drag&Drop
Procedure TFormMain.WMDropFiles(Var Msg : TMessage);
Const BuffSize = 255;
Var FileName : Array[0..Buffsize] of Char;
  n, i : integer;
  Fn : String;
Begin
  Try
    n := DragQueryFile(Msg.wParam, $FFFFFFFF, Nil, BuffSize); // ドラッグファイル数を取り出す
    For i := 0 To n-1 Do begin
      DragQueryFile(Msg.wParam, i, FileName, BuffSize);
      Fn := String(FileName); // ドラッグされたファイル名を取り出す
      // Fn に対する処理
      LabeledEditFileName.Text := Fn;
    End;
    Msg.Result := 0;
    if LowerCase(ExtractFileExt(LabeledEditFileName.Text)) = '.lnk' then
      LabeledEditFileName.Text := GetLnkFname(LabeledEditFileName.Text);
    if Trim(LabeledEditFileName.Text)<>'' then DispInfo(LabeledEditFileName.Text) else begin
      ListView1.Clear;
    end;
  Finally
    DragFinish(Msg.wParam);
  End;
End;


// StringGridのテキストをクリップボードにコピーする
procedure CopyClip(Ch : Char);
Var St : String; n : Integer;
begin
  With FormMain do begin
    if ListView1.Items.Count>0 then begin
      St := '['+ListView1.Columns.Items[0].Caption+']'+ Ch + '['+ListView1.Columns.Items[1].Caption+']'+^M +^J;
      For n := 0 to ListView1.Items.Count -1 do
        if ListView1.Items[n].Caption<>'' then
          St := St + ListView1.Items[n].Caption + Ch + ListView1.Items[n].SubItems[0] +^M +^J;
      Clipboard.AsText := St;
    end
    else Clipboard.AsText := '';
  end;
end;


// CheckBoxBlank クリックイベント
procedure TFormMain.CheckBoxBlankClick(Sender: TObject);
begin
  if Trim(LabeledEditFileName.Text)<>'' then DispInfo(LabeledEditFileName.Text);
end;


// クリップボードへコピー TAB区切り
procedure TFormMain.ButtonCopyTabClick(Sender: TObject);
begin
  CopyClip(^I);
end;


// クリップボードへコピー コンマ区切り
procedure TFormMain.ButtonCopyComClick(Sender: TObject);
begin
  CopyClip(',');
end;


// キー入力イベント
procedure TFormMain.LabeledEditFileNameKeyPress(Sender: TObject; var Key: Char);
begin
  if Key=^M then begin
    Key := #0;
    DispInfo(LabeledEditFileName.Text);
  end;
end;


// フォームリサイズ・イベント
procedure TFormMain.FormResize(Sender: TObject);
begin
  LabeledEditFileName.Width := FormMain.ClientWidth - LabeledEditFileName.Left - ButtonOpen.Width - 20;
  ListView1.Width    := LabeledEditFileName.Width;
  ButtonOpen.Left    := LabeledEditFileName.Left + LabeledEditFileName.Width + 10;
  CheckBoxBlank.Left := ButtonOpen.Left;
  ButtonExit.Left    := ButtonOpen.Left;
  ButtonCopyCom.Left := ButtonOpen.Left;
  ButtonCopyTab.Left := ButtonOpen.Left;
  ListView1.Height   := FormMain.ClientHeight - ListView1.Top -10;
  ButtonExit.Top     :=  ListView1.Top + ListView1.Height - ButtonExit.Height;
  ButtonCopyCom.Top  := ButtonExit.Top -40;
  ButtonCopyTab.Top  := ButtonCopyCom.Top -30;
end;


// フォーム作成イベント
procedure TFormMain.FormCreate(Sender: TObject);
begin
  Caption := 'Version informasion  Ver'+ GetFxeVersion;
 {$IFDEF DEBUG} Caption := Caption + ' *** Debug mode ***'; {$ENDIF}
  ListView1.Clear;
  DragAcceptFiles(Handle, True);  // for Drag&Drop
  ListView1.Clear;
  // Hint表示を設定する
  if GetUserDefaultLCID <> $411 then begin // 日本語以外
    LabeledEditFileName.Hint := 'Enter the file name with the full path and press the "Enter" key';
    ButtonOpen.Hint     := 'Launch an Open dialog';
    CheckBoxBlank.Hint  := 'Display even blank items';
    ButtonCopyTab.Hint  := 'Copy to clipboard separated by Tabs';
    ButtonCopyCom.Hint  := 'Copy to clipboard separated by Comma';
    ButtonExit.Hint     := 'Terminate';
  end
  else begin
    LabeledEditFileName.Hint := 'ファイル名をフルパスで入力して「Enter」キーを押す';
    ButtonOpen.Hint     := 'オープンダイアログを起動する';
    CheckBoxBlank.Hint  := '内容が空白の項目も表示する';
    ButtonCopyTab.Hint  := 'タブ区切りでクリップボードへコピーする';
    ButtonCopyCom.Hint  := 'コンマ区切りでクリップボードへコピーする';
    ButtonExit.Hint     := '終了する';
  end;
end;


// 終了
procedure TFormMain.ButtonExitClick(Sender: TObject);
begin
  Close;
end;


end.
